(define-package "elpher" "3.7.0" "A friendly gopher and gemini client"
  '((emacs "27.1"))
  :keywords ("convenience")
  :authors (("Tim Vaughan" . "plugd@thelambdalab.xyz"))
  :maintainer ("Tim Vaughan" . "plugd@thelambdalab.xyz")
  :url "https://thelambdalab.xyz/elpher")
;; Local Variables:
;; no-byte-compile: t
;; end:
